from clease.db_util import (get_all_cf_names, get_all_cf, get_cf_tables)
import argparse


def db_cli(args: argparse.Namespace):
    if args.show == 'tab':
        show_cf_tables(args.name)
    elif args.show == 'names':
        show_cf_names(args.name)
    elif args.show == 'cf':
        try:
            show_cf(args.name, args.id)
        except Exception as exc:
            print(exc)


def show_cf_tables(db_name: str):
    tables = get_cf_tables(db_name)
    cftab = ', '.join(tables)
    print(f"Available correlation function tables: {cftab}")


def show_cf_names(db_name: str):
    table = get_cf_tables(db_name)[0]
    names = sorted(get_all_cf_names(db_name, table))
    print("Tracked correlation functions")

    line = ""
    char_per_line = 110
    for name in names:
        if len(line) + len(name) > char_per_line:
            print(line)
            line = ""

        if line == "":
            line += name
        else:
            line += ', ' + name
    print(line)


def show_cf(db_name: str, db_id: int):
    table = get_cf_tables(db_name)[0]
    cfs = get_all_cf(db_name, table, db_id)
    print(f"Correlation functions for ID {db_id}")
    for k, v in cfs.items():
        print(f"{k}: {v}")



from kivy.uix.screenmanager import Screen
from kivy.uix.textinput import TextInput
from kivy.uix.label import Label
from kivy.uix.stacklayout import StackLayout
from kivy.uix.spinner import Spinner
from kivy.uix.button import Button
from kivy.app import App

from clease.gui.util import parse_grouped_basis_elements, parse_elements
import traceback


class ConcentrationPage(Screen):
    next_constraint_id = 0
    elements = []
    grouped_elements = []
    grouped_basis = None

    def _elements_changed(self, new_elements):
        if len(self.grouped_elements) != len(new_elements):
            return True

        if any(len(a) != len(b) for a, b in zip(self.grouped_elements, new_elements)):
            return True

        return any(
            any(symb1 != symb2
                for symb1, symb2 in zip(a, b))
            for a, b in zip(self.grouped_elements, new_elements))

    def _group_elements(self, elements, grouped_basis):
        if grouped_basis is None:
            return elements

        gr_elements = []
        for group in grouped_basis:
            gr_elements.append(elements[group[0]])
        return gr_elements

    @property
    def num_conc_vars(self):
        return sum(len(item) for item in self.grouped_elements)

    def add_constraint(self):
        layout = StackLayout(id=f"cnst{self.next_constraint_id}", size_hint=[1, 0.05])
        width = 1.0 / float((self.num_conc_vars + 3))
        for i in range(self.num_conc_vars):
            layout.add_widget(
                TextInput(text='0',
                          size_hint=[width, 1],
                          write_tab=False,
                          multiline=False,
                          id=f"conc{i}"))
        layout.add_widget(
            Spinner(text='<=',
                    values=['<=', '>=', '='],
                    id='comparisonSpinner',
                    size_hint=[width, 1]))
        layout.add_widget(
            TextInput(text='0', size_hint=[width, 1], write_tab=False, multiline=False, id='rhs'))
        layout.add_widget(
            Button(text='Remove',
                   size_hint=[width, 1],
                   on_press=lambda _: self.remove_constraint(layout)))

        self.ids.mainConcLayout.add_widget(layout)
        self.next_constraint_id += 1
        return layout

    def remove_constraint(self, widget):
        self.ids.mainConcLayout.remove_widget(widget)

    def check_user_input(self):
        elems = self.ids.elementInput.text
        if elems == '':
            msg = 'No elements are given.'
            App.get_running_app().root.ids.status.text = msg
            return 1

        if not self.elem_ok():
            return 1

        gr_basis = self.ids.groupedBasisInput.text
        if gr_basis != '':
            if not self.grouped_basis_ok():
                return 1

        for widget in self.ids.mainConcLayout.children:
            if widget.id is None:
                continue

            if widget.id.startswith('cnst'):
                for child in widget.children:
                    if child.id is None:
                        continue
                    if child.id.startswith('conc') or child.id == 'rhs':
                        try:
                            _ = float(child.text)
                        except Exception:
                            traceback.print_exc()
                            msg = "All constraints need to be float"
                            App.get_running_app().root.ids.status.text = msg
                            return 1

        return 0

    def get_constraint_matrices(self):
        if self.check_user_input() != 0:
            raise ValueError("Could not parse elements or constraints")

        # Loop across widgets
        A_lb = []
        A_eq = []
        rhs_lb = []
        rhs_eq = []
        for widget in self.ids.mainConcLayout.children:
            if widget.id is None:
                continue

            if widget.id.startswith('cnst'):
                matrix_row, rhs, equality = self._extract_row(widget)

                if equality:
                    A_eq.append(matrix_row)
                    rhs_eq.append(rhs)
                else:
                    A_lb.append(matrix_row)
                    rhs_lb.append(rhs)
        return A_lb, rhs_lb, A_eq, rhs_eq

    def _clear_constraints(self, clear_elemHeader=True):
        for child in self.ids.mainConcLayout.children[:]:
            if child.id is None:
                continue
            if child.id.startswith('cnst'):
                self.ids.mainConcLayout.remove_widget(child)
            elif child.id == 'elemHeader' and clear_elemHeader:
                self.ids.mainConcLayout.remove_widget(child)

    def _extract_row(self, widget):
        row = []
        ids = []
        rhs = 0.0
        spinner_value = ''
        for child in widget.children:
            if child.id == 'comparisonSpinner':
                spinner_value = child.text

        if spinner_value == '':
            raise ValueError("Could not extract the spinner value")

        equality = spinner_value == '='
        sign = 1.0
        if spinner_value == '<=':
            sign = -1.0
        for child in widget.children:
            if child.id is None:
                continue

            if child.id.startswith('conc'):
                row.append(sign * float(child.text))
                ids.append(child.id)
            elif child.id == 'rhs':
                rhs = sign * float(child.text)

        zipped = sorted(zip(ids, row))
        row = [z[1] for z in zipped]
        return row, rhs, equality

    def apply_Elements_GroupedBasis(self):
        elem_str = self.ids.elementInput.text
        grouped_basis = self.ids.groupedBasisInput.text
        try:
            if grouped_basis != '':
                self.grouped_basis = \
                    parse_grouped_basis_elements(grouped_basis)

            elements = parse_elements(elem_str)
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)
            return
        new_elements = self._group_elements(elements, self.grouped_basis)

        if self._elements_changed(new_elements):
            self._clear_constraints()

            self.grouped_elements = new_elements
            self.elements = elements

            layout = StackLayout(id='elemHeader', size_hint=[1, 0.10])
            width = 1.0 / float(self.num_conc_vars + 3)
            layout.add_widget(Label(text='', size_hint=[1, 0.5]))

            for item in self.grouped_elements:
                for sym in item:
                    layout.add_widget(Label(text=sym, size_hint=[width, 0.5]))

            layout.add_widget(Label(text='Type', size_hint=[width, 0.5]))
            layout.add_widget(Label(text='RHS', size_hint=[width, 0.5]))
            layout.add_widget(Label(text='', size_hint=[width, 0.5]))
            self.ids.mainConcLayout.add_widget(layout)

        msg = "Applied Elements and Grouped basis fields."
        App.get_running_app().root.ids.status.text = msg

    def todict(self):
        A_lb, rhs_lb, A_eq, rhs_eq = self.get_constraint_matrices()
        data = {
            'elements': self.ids.elementInput.text,
            'grouped_basis': self.ids.groupedBasisInput.text,
            'A_lb': A_lb,
            'rhs_lb': rhs_lb,
            'A_eq': A_eq,
            'rhs_eq': rhs_eq
        }
        return data

    def set_Elements_GroupedBasis(self, elements, grouped_basis):
        self.ids.elementInput.text = elements
        self.ids.groupedBasisInput.text = grouped_basis
        self.apply_Elements_GroupedBasis()

    def load_from_matrices(self, A_lb, rhs_lb, A_eq, rhs_eq):
        # remove constraints if there are any
        self._clear_constraints(clear_elemHeader=False)

        for i in range(len(A_lb)):
            layout = self.add_constraint()

            # Traverse the children
            for child in layout.children:
                if child.id is None:
                    continue
                if child.id.startswith('conc'):
                    col = int(child.id[-1])
                    child.text = str(abs(A_lb[i][col]))
                elif child.id == 'rhs':
                    child.text = str(abs(rhs_lb[i]))
                elif child.id == 'comparisonSpinner':
                    if rhs_lb[i] < 0.0:
                        child.text = '<='
                    else:
                        child.text = '>='

        for i in range(len(A_eq)):
            layout = self.add_constraint()

            # Traverse children
            for child in layout.children:
                if child.id is None:
                    continue
                if child.id.startswith('conc'):
                    col = int(child.id[-1])
                    child.text = str(A_eq[i][col])
                elif child.id == 'rhs':
                    child.text = str(abs(rhs_eq[i]))
                elif child.id == 'comparisonSpinner':
                    child.text = '='

    def elem_ok(self):
        elems = self.ids.elementInput.text
        try:
            _ = parse_elements(elems)
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)
            return False
        return True

    def grouped_basis_ok(self):
        gr_basis = self.ids.groupedBasisInput.text
        try:
            _ = parse_grouped_basis_elements(gr_basis)
        except Exception as exc:
            traceback.print_exc()
            App.get_running_app().root.ids.status.text = str(exc)
            return False
        return True



# -*- coding: utf-8 -*-
"""Monte Carlo method for ase."""
import sys
import datetime
import time
import logging
from collections import Counter
from typing import Sequence
import numpy as np
from ase import Atoms
from ase.units import kB
from clease.tools import SystemChange
from clease.montecarlo.exponential_filter import ExponentialFilter
from clease.montecarlo.averager import Averager
from clease.montecarlo.bias_potential import BiasPotential
from clease.montecarlo.observers import MCObserver
from clease.montecarlo.trial_move_generator import TrialMoveGenerator, RandomSwap

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class Montecarlo:
    """Class for running Monte Carlo at a fixed composition (canonical).

    :param atoms: ASE Atoms object (with CE calculator attached)
    :param temp: Temperature of Monte Carlo simulation in Kelvin
    :param generator: Generator that produces trial moves
    """

    def __init__(self, atoms: Atoms, temp: float, generator: TrialMoveGenerator = None):
        self.name = "MonteCarlo"
        self.atoms = atoms
        self.T = temp

        if generator is None:
            self.generator = RandomSwap(atoms)
        else:
            self.generator = generator

        # List of observers that will be called every n-th step
        # similar to the ones used in the optimization routines
        self.observers = []
        self.bias_potentials = []

        self.current_step = 0
        self.num_accepted = 0
        self.status_every_sec = 30
        E0 = self.atoms.calc.calculate(None, None, None)
        self.atoms.calc.clear_history()
        self.current_energy = E0
        self.new_energy = self.current_energy

        self.trial_move = []  # Last trial move performed
        self.mean_energy = Averager(ref_value=self.current_energy)
        self.energy_squared = Averager(ref_value=self.current_energy)

        # Some member variables used to update the atom tracker, only relevant
        # for canonical MC
        self.quit = False

        self.filter = ExponentialFilter(min_time=0.2 * len(self.atoms),
                                        max_time=20 * len(self.atoms),
                                        n_subfilters=10)

    def update_current_energy(self):
        """Enforce a new energy evaluation."""
        self.current_energy = self.atoms.get_potential_energy()

    def reset(self):
        """Reset all member variables to their original values."""
        logger.debug('Resetting.')
        for _, obs in self.observers:
            obs.reset()

        self.filter.reset()
        self._reset_internal_counters()

        self.mean_energy.clear()
        self.energy_squared.clear()

    def _reset_internal_counters(self):
        """Reset the step counters which are used internally"""
        self.current_step = 0
        self.num_accepted = 0

    def add_bias(self, potential: BiasPotential):
        """Add a new bias potential.

        Parameters:

        potential:
            Potential to be added
        """
        if not isinstance(potential, BiasPotential):
            raise TypeError("potential has to be of type BiasPotential")
        self.bias_potentials.append(potential)

    def attach(self, obs: MCObserver, interval: int = 1):
        """Attach observers to be called on a given MC step interval.

        Parameters:

        obs: MCObserver
            Observer to be added

        interval: int
            How often the observer should be called
        """
        if not obs.interval_ok(interval):
            name = type(obs).__name__
            raise ValueError(f"Invalid interval for {name}. Check docstring of the observer.")

        self.observers.append((interval, obs))

    def _initialize_run(self):
        self.generator.initialize(self.atoms)

        self.update_current_energy()
        # Reset the internal step counters
        self._reset_internal_counters()
        # Atoms object should have attached calculator
        # Add check that this is show
        self._mc_step()

    def run(self, steps: int = 100):
        """Run Monte Carlo simulation.

        Parameters:

        steps: int
            Number of steps in the MC simulation
        """
        # Check the number of different elements are correct to avoid
        # infinite loops
        self._initialize_run()

        start = time.time()
        prev = self.current_step
        while self.current_step < steps:
            E, _ = self._mc_step()

            self.mean_energy += E
            self.energy_squared += E**2

            if time.time() - start > self.status_every_sec:
                ms_per_step = 1000.0 * self.status_every_sec / (self.current_step - prev)
                accept_rate = self.num_accepted / self.current_step
                logger.info("%d of %d steps. %.2f ms per step. Acceptance rate: %.2f",
                            self.current_step, steps, ms_per_step, accept_rate)
                prev = self.current_step
                start = time.time()

            if self.quit:
                logger.debug('Quit signal detected. Breaking.')
                break

        logger.info("Reached maximum number of steps (%d mc steps)", steps)

    @property
    def meta_info(self):
        """Return dict with meta info."""
        ts = time.time()
        st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')
        v_info = sys.version_info
        meta_info = {
            "timestamp": st,
            "python_version": f"{v_info.major}.{v_info.minor}.{v_info.micro}"
        }
        return meta_info

    def get_thermodynamic_quantities(self):
        """Compute thermodynamic quantities."""
        quantities = {}
        mean_energy = self.mean_energy.mean
        quantities["energy"] = mean_energy
        mean_sq = self.energy_squared.mean
        quantities["heat_capacity"] = \
            (mean_sq - mean_energy**2) / (kB * self.T**2)
        quantities["energy_var"] = mean_sq - mean_energy**2
        quantities["temperature"] = self.T
        at_count = self.count_atoms()
        for key, value in at_count.items():
            name = f"{key}_conc"
            conc = float(value) / len(self.atoms)
            quantities[name] = conc

        # Add some more info that can be useful
        quantities.update(self.meta_info)

        # Add information from observers
        for obs in self.observers:
            quantities.update(obs[1].get_averages())
        return quantities

    def _calculate_step(self, system_changes: Sequence[SystemChange]):
        """Calculate energies given a step, and decide if we accept the step.

        Returns boolean if system changes are accepted.

        Parameters:

        system_changes: list
            List with system changes
        """
        # NOTE: Calculate updates the system
        calc = self.atoms.calc

        with calc.with_system_changes(system_changes) as keeper:
            self.new_energy = calc.get_energy()

            # NOTE: As this is called after calculate, the changes has
            # already been introduced to the system
            for bias in self.bias_potentials:
                self.new_energy += bias(system_changes)

            logger.debug('Current energy: %.3f eV, new energy: %.3f eV', self.current_energy,
                         self.new_energy)
            accept = self._do_accept(self.current_energy, self.new_energy)
            logger.debug('Change accepted? %s', accept)

            # Decide if we keep changes, or rollback
            keeper.keep_changes = accept
        return accept

    def _do_accept(self, current_energy: float, new_energy: float) -> bool:
        """Decide if we accept a state, based on the energies.

        Return a bool on whether the move was accepted.

        Parameters:

        :param current_energy: Energy of the current configuration
        :param new_energy: Energy of the new configuration.
        """
        # Standard Metropolis acceptance criteria
        if new_energy < current_energy:
            return True
        kT = self.T * kB
        energy_diff = new_energy - current_energy
        probability = np.exp(-energy_diff / kT)
        logger.debug('Calculated probability: %.3f', probability)
        return np.random.rand() <= probability

    def _move_accepted(self, system_changes: Sequence[SystemChange]) -> Sequence[SystemChange]:
        logger.debug('Move accepted, updating things')
        self.num_accepted += 1
        self.generator.on_move_accepted(system_changes)
        self.current_energy = self.new_energy
        return system_changes

    # pylint: disable=no-self-use
    def _move_rejected(self, system_changes: Sequence[SystemChange]) -> Sequence[SystemChange]:
        logger.debug('Move rejected, undoing system changes: %s', system_changes)
        self.generator.on_move_rejected(system_changes)

        # Move rejected, no changes are made
        system_changes = [
            SystemChange(index=change.index,
                         old_symb=change.old_symb,
                         new_symb=change.old_symb,
                         name=change.name) for change in system_changes
        ]
        logger.debug('Reversed system changes: %s', system_changes)
        return system_changes

    def count_atoms(self):
        """Count the number of each element."""
        return dict(Counter(self.atoms.symbols))

    def _mc_step(self):
        """Make one Monte Carlo step by swithing two atoms."""
        self.current_step += 1

        system_changes = self.generator.get_trial_move()
        self.trial_move = system_changes

        # Calculate step, and whether we accept it
        move_accepted = self._calculate_step(system_changes)

        updater = self._move_accepted if move_accepted else self._move_rejected
        system_changes = updater(system_changes)

        # Execute all observers
        self.execute_observers(system_changes)
        self.filter.add(self.current_energy)
        return self.current_energy, move_accepted

    def execute_observers(self, system_changes: Sequence[SystemChange]):
        for interval, obs in self.observers:
            if self.current_step % interval == 0:
                logger.debug('Executing observer %s at step %d with interval %d.', obs.name,
                             self.current_step, interval)
                obs(system_changes)



from typing import Sequence, Dict
from clease.tools import SystemChange
from clease.montecarlo.observers.mc_observer import MCObserver

__all__ = ('AcceptanceRate',)


class AcceptanceRate(MCObserver):
    """
    Observer that tracks the fraction of monte carlo steps that is accepted
    """

    def __init__(self):
        super().__init__()
        self.num_calls = 0
        self.num_accept = 0

    @property
    def rate(self) -> float:
        """
        Acceptance rate
        """
        if self.num_calls == 0:
            return 0.0
        return self.num_accept / self.num_calls

    def reset(self):
        """
        Reset the observer
        """
        self.num_calls = 0
        self.num_accept = 0

    def __call__(self, system_changes: Sequence[SystemChange]) -> None:
        """
        Updates the acceptance rate.

        :param system_changes: Sequence with introduces changes. A move is considered
            accepted if old_symb != new_symb in any of the changes
        """
        accepted = any(c.old_symb != c.new_symb for c in system_changes)

        if accepted:
            self.num_accept += 1
        self.num_calls += 1

    def get_averages(self) -> Dict[str, float]:
        """
        Return dictionary with the rate such that it is added to thermodynaic quantities
        """
        return {'accept_rate': self.rate}



from typing import List, Tuple, Sequence
import warnings
import logging
import time
from ase import Atoms
from ase.units import kB
import numpy as np
from clease.montecarlo import BarrierModel
from clease.montecarlo import KMCEventType
from clease.montecarlo.observers import MCObserver
from clease.tools import SystemChange

__all__ = ('KineticMonteCarlo',)

logger = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class KineticMonteCarlo:
    """
    Kinetic Monte Carlo using the residence time algorithm, which is a
    rejection free algorithm and is thus also efficient at low temperatures.

    References:

    [1] Bortz, Alfred B., Malvin H. Kalos, and Joel L. Lebowitz.
        "A new algorithm for Monte Carlo simulation of Ising spin systems."
        Journal of Computational Physics 17.1 (1975): 10-18.

    [2] Rautiainen, T. T., and A. P. Sutton.
        "Influence of the atomic diffusion mechanism on morphologies,
        kinetics, and the mechanisms of coarsening during phase separation."
        Physical Review B 59.21 (1999): 13681.


    :param atoms: Atoms object where KMC should be executed
    :param barrier: Model used to evaluate the barriers
    :param events: List of KMCEvents used to produce possible events
    """

    def __init__(self, atoms: Atoms, T: float, barrier: BarrierModel,
                 event_types: Sequence[KMCEventType]):
        self.atoms = atoms
        self.T = T
        self.barrier = barrier
        self.event_types = event_types
        self.time = 0.0
        self.attempt_freq = 1e4
        self.observers = []
        self.log_interval = 30
        self.epr = None

    def reset(self):
        """
        Reset the KMC class. All information stored as attributes of KMC and
        in observers will be cleared.
        """
        self.time = 0.0
        for _, obs in self.observers:
            obs.reset()

        if self.epr is not None:
            self.epr.reset()

    @property
    def kT(self) -> float:
        return kB * self.T

    def _update_epr(self, current: int, choice: int, swaps: List[int], cum_rates: np.ndarray):
        """
        Update the EPR tracker (if set).

        :param current: Current position of the vacancy
        :param choice: Chosen index in the cum_rates array
        :param swaps: List of possible swaps
        :param cum_rates: Cummulative transition rates
        """
        if self.epr is None:
            return
        self.epr.update(current, choice, cum_rates, swaps)

    def attach(self, obs: MCObserver, interval: int = 1):
        """
        Attach an observer to the MC run

        :param obs: Instance of MCObserver
        :param interval: Number of steps between each time the observer
            is executed.
        """
        self.observers.append((interval, obs))

    def _rates(self, vac_idx: int) -> Tuple[List[int], np.ndarray]:
        """
        Return the rates for all possible swaps for a vacancy
        at position vac_idx

        :param vac_idx: Position of the vacancy
        """
        swaps = [
            swap for event in self.event_types for swap in event.get_swaps(self.atoms, vac_idx)
        ]
        if len(swaps) == 0:
            raise RuntimeError("No swaps are possible.")
        rates = [self._get_rate_from_swap(swap, vac_idx) for swap in swaps]
        return swaps, np.array(rates)

    def _get_rate_from_swap(self, swap_idx: int, vac_idx: int) -> float:
        symb = self.atoms[swap_idx].symbol
        system_change = [
            SystemChange(index=vac_idx, old_symb='X', new_symb=symb, name='kmc_swap'),
            SystemChange(index=swap_idx, old_symb=symb, new_symb='X', name='kmc_swap')
        ]
        Ea = self.barrier(self.atoms, system_change)
        rate = self.attempt_freq * np.exp(-Ea / self.kT)
        if rate < 0.0:
            warnings.warn("Negative rate encountered!")
            rate = 0.001
        return rate

    def _mc_step(self, vac_idx: int, step_no: int) -> int:
        """
        Perform an MC step and return the new index of the moving vacancy

        :param vac_idx: Current position of the vacancy
        :param step_no: Step number
        """
        swaps, rates = self._rates(vac_idx)
        tau = 1.0 / np.sum(rates)
        rates *= tau
        cumulative_rates = np.cumsum(rates)

        rnd = np.random.random()

        # Argmax returns the first occurence of True
        cum_indx = np.argmax(cumulative_rates > rnd)
        choice = swaps[cum_indx]
        self._update_epr(vac_idx, cum_indx, swaps, cumulative_rates)

        # Apply step
        symb = self.atoms[choice].symbol
        system_change = [
            SystemChange(index=vac_idx, old_symb='X', new_symb=symb, name='kmc_swap'),
            SystemChange(index=choice, old_symb=symb, new_symb='X', name='kmc_swap')
        ]

        # Trigger update
        self.atoms.calc.update_cf(system_change)
        self.atoms.calc.clear_history()
        self.time += tau

        for interval, obs in self.observers:
            if step_no % interval == 0:
                obs(system_change)
        return choice

    def run(self, num_steps: int, vac_idx: int):
        """
        Run a given number of MC steps. The algorithm uses only
        one vacancy that moves around. Note that there can be more
        vacancies in the system, but they are just treated as regular
        atoms.

        :param num_steps: Number of steps to perform
        :param vac_idx: Index where the moving vacancy is located
        """

        if self.atoms[vac_idx].symbol != 'X':
            raise ValueError(f"Index {vac_idx} is not a vacancy. "
                             f"Symbol: {self.atoms[vac_idx].symbol}")

        # Trigger one energy evaluation to make sure that CFs are in sync
        self.atoms.get_potential_energy()
        self.atoms.calc.clear_history()

        now = time.time()
        for i in range(num_steps):
            if self.log_interval is not False and time.time() - now > self.log_interval:
                logger.info("Step %d of %d", i, num_steps)
                now = time.time()
            vac_idx = self._mc_step(vac_idx, i)

        # If entropy is tracked: Flush the buffer
        if self.epr is not None:
            self.epr.flush_buffer()



from deprecated import deprecated
from clease import settings

__all__ = ('Concentration',)


@deprecated(version='0.10.0', reason='Use Concentration class from clease.settings instead')
class Concentration(settings.Concentration):
    """Concentration class was moved to clease.settings"""



# pylint: disable=undefined-variable
import logging
from deprecated import deprecated
from .version import *
from .template_filters import *
from .atoms_manager import *
from .convexhull import *
from .data_manager import *
from .structure_mapper import *
from .evaluate import *
from .structgen.new_struct import NewStructures  # Promotion for convenience
from .svd import *
from .concentration import *
from .regression_old import *
from .cluster_coverage import *

from . import cluster
from . import structgen
from . import corr_func
from . import settings
from . import basis_function
from . import montecarlo
from . import mp_logger
from . import tools
from . import template_atoms
from . import concentration
from . import db_util
from . import jsonio
# Import this global, so it can be disabled with "clease.REQUIRE_COMPATIBLE_TABLE_VERISON = False"
from .db_util import toggle_require_compatible_table_version

logger = logging.getLogger(__name__)


@deprecated(version='0.10.0', reason='import CEBulk from clease.settings instead')
def CEBulk(*args, **kwargs):
    return settings.CEBulk(*args, **kwargs)


@deprecated(version='0.10.0', reason='import CECrystal from clease.settings instead')
def CECrystal(*args, **kwargs):
    return settings.CECrystal(*args, **kwargs)


@deprecated(version='0.10.0', reason='import CESlab from clease.settings instead')
def CESlab(*args, **kwargs):
    return settings.CESlab(*args, **kwargs)


ADDITIONAL = ('settings', 'basis_function', 'jsonio', 'corr_func', 'structgen', 'NewStructures',
              'cluster', 'tools', 'montecarlo', 'mp_logger', 'template_atoms', 'CEBulk',
              'CECrystal', 'CESlab', 'concentration', 'db_util',
              'toggle_require_compatible_table_version')

__all__ = (version.__all__ + template_filters.__all__ + atoms_manager.__all__ +
           concentration.__all__ + evaluate.__all__ + convexhull.__all__ + data_manager.__all__ +
           structure_mapper.__all__ + svd.__all__ + regression_old.__all__ +
           cluster_coverage.__all__) + ADDITIONAL



from typing import Sequence
from ase.atoms import Atoms
from ase.io.trajectory import TrajectoryWriter
from clease.tools import SystemChange
from clease.montecarlo.observers import MCObserver
from clease.tools import add_file_extension


class Snapshot(MCObserver):
    """Store a snapshot in a trajectory file.

    :param atoms: Instance of the atoms objected modofied by the MC object
    :param fname: Name of the trajectory file. Adds extension '.traj' if none is given.
    :param mode: IO mode used by the ASE TrajectoryWriter (must be w or a)
    """

    name = "Snapshot"

    def __init__(self, atoms: Atoms, fname: str = "snapshot.traj", mode: str = "w"):
        super().__init__()
        full_fname = add_file_extension(fname, '.traj')
        self.atoms = atoms
        self.traj = TrajectoryWriter(full_fname, mode=mode)
        self.fname = full_fname

    def __call__(self, system_changes: Sequence[SystemChange]):
        """Write a snapshot to a .traj file.

        Parameters:

        system_changes: list
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        self.traj.write(self.atoms)



from typing import Sequence
import numpy as np
from clease.tools import SystemChange
from clease.montecarlo import BiasPotential


class BinnedBiasPotential(BiasPotential):
    """
    Binned bias potential is a bias potential that is sampled on a grid.

    Parameters:

    xmin: float
        Minimum value for the collective variable

    xmax: float
        Maximum value for the collective variable

    nbins: int
        Number of bins

    getter: MCObserver
        Callable object for getting the collective variable.
        It needs to support the peak key word.
    """

    def __init__(self, xmin=0.0, xmax=1.0, nbins=10, getter=None):
        self.xmin = xmin
        self.xmax = xmax
        self.nbins = nbins
        self.values = np.zeros(self.nbins)
        self.dx = (self.xmax - self.xmin) / self.nbins
        self.getter = getter

    def get_index(self, x):
        """Return the index corresponding to an x value."""
        i = int((x - self.xmin) / self.dx)
        if i >= self.nbins:
            i = self.nbins - 1
        elif i < 0:
            i = 0
        return i

    def get_x(self, index):
        """Return x value corresponding to an index."""
        return self.xmin + index * self.dx + 0.5 * self.dx

    def evaluate(self, x):
        """Evaluate the bias potential at x."""
        i = self.get_index(x)
        if i == 0:
            y_left = self.values[0]
            y_right = self.values[1]
            x_left = self.get_x(0)
            y = y_left + (x - x_left) * (y_right - y_left) / self.dx
        elif i == self.nbins - 1:
            y_left = self.values[self.nbins - 2]
            y_right = self.values[self.nbins - 1]
            x_left = self.get_x(self.nbins - 2)
            y = y_left + (x - x_left) * (y_right - y_left) / self.dx
        else:
            x_left = self.get_x(i - 1)
            x_center = x_left + self.dx
            x_right = x_center + self.dx
            y_left = self.values[i - 1]
            y_center = self.values[i]
            y_right = self.values[i + 1]
            fac = 1.0 / self.dx**2

            # Lagrange polynomial
            y = 0.5*fac*(x - x_center)*(x-x_right)*y_left - \
                fac*(x - x_left)*(x - x_right)*y_center + \
                0.5*fac*(x - x_left)*(x-x_center)*y_right
        return y

    def __call__(self, system_changes: Sequence[SystemChange]):
        """Get the bias potential value after the system_change."""
        x = self.getter(system_changes, peak=True)
        return self.evaluate(x)

    def todict(self):
        """Return dictionary representation. (Does not include the getter)"""
        return {
            'xmin': self.xmin,
            'xmax': self.xmax,
            'values': self.values.tolist(),
            'nbins': self.nbins,
            'dx': self.dx
        }

    def from_dict(self, data):
        """Initialize from a dictionary (Does not include the getter)."""
        self.xmin = data['xmin']
        self.xmax = data['xmax']
        self.values = np.array(data['values'])
        self.nbins = data['nbins']
        self.dx = data['dx']

    def calculate_from_scratch(self, atoms):
        """Calculate the value from scratch."""
        x = self.getter.calculate_from_scratch(atoms)
        return self.evaluate(x)

    def local_update(self, x, dE):
        """Perform a local update."""
        i = self.get_index(x)
        self.values[i] += dE

    def zero(self):
        """Set all values to zero."""
        self.values[:] = 0

    def get_coeff(self):
        """Return all values."""
        return self.values



from typing import List, Tuple
from ase import Atoms


class MCObserver:
    """Base class for all MC observers."""
    name = "GenericObserver"

    def __call__(self, system_changes: List[Tuple[int, str, str]]):
        """
        Gets information about the system changes and can perform some action

        :param system_changes: List of system changes. For example, if the
            occupation of the atomic index 23 is changed from Mg to Al,
            system_change = [(23, Mg, Al)]. If an Mg atom occupying the atomic
            index 26 is swapped with an Al atom occupying the atomic index 12,
            system_change = [(26, Mg, Al), (12, Al, Mg)]
        """
        pass

    def reset(self) -> None:
        """Reset all values of the MC observer"""
        pass

    def get_averages(self) -> dict:
        """Return averages in the form of a dictionary."""
        return {}

    def calculate_from_scratch(self, atoms: Atoms) -> None:
        """
        Method for calculating the tracked value from scratch
        (i.e. without using fast update methods)
        """
        pass

    def interval_ok(self, interval: int) -> bool:
        """
        Check if the interval specified on attach is ok. Default is that all
        intervals are OK

        :param interval: Interval controlling how often a MC observer will be
            called.
        """
        return True



from clease.montecarlo.constraints import MCConstraint


class FixedElement(MCConstraint):
    """
    Class for forcing an element of a certiain type to stay fixed.

    Parameters:

    element: str
        Name of the element that is supposed to stay fixed
    """

    def __init__(self, element):
        self.element = element

    def __call__(self, system_changes):
        for change in system_changes:
            if change[1] == self.element or change[2] == self.element:
                return False
        return True
